#include "defaultplayercontroller.h"

#include "internal.h"
#include "internal/physics.h"

#include <glm/gtx/vector_angle.hpp>
#include <BulletCollision/CollisionDispatch/btCollisionWorld.h>

namespace kek {

static bool castPlayer(glm::vec3 pos, glm::vec3 delta, const btCollisionObject **outBody, float *outFrac, glm::vec3 *outPoint, glm::vec3 *outNormal){
	btTransform from;
	from.setIdentity();
	from.setOrigin(Physics::fromGLM(pos));

	btTransform to;
	to.setIdentity();
	to.setOrigin(Physics::fromGLM(pos + delta));

	ClosestNotSelfConvexResultCallback cb = ClosestNotSelfConvexResultCallback(from.getOrigin(), to.getOrigin(), kekData.player->physics->body);
	cb.m_collisionFilterGroup = -1;
	cb.m_collisionFilterMask = -1;
	kekData.physics->world->convexSweepTest((btConvexShape *) kekData.player->physics->body->getCollisionShape(), from, to, cb);
	if(!cb.m_hitCollisionObject) return false;

	*outBody = cb.m_hitCollisionObject;
	*outFrac = cb.m_closestHitFraction;
	*outPoint = Physics::toGLM(cb.m_hitPointWorld);
	*outNormal = Physics::toGLM(cb.m_hitNormalWorld);
	return true;
}

DefaultPlayerController::DefaultPlayerController() {

}

DefaultPlayerController::~DefaultPlayerController() {

}

glm::vec3 DefaultPlayerController::move(glm::vec3 movement) {
	glm::vec3 pos = kekData.player->getPosition();

	int count = 0;
	while (count < 3) {
		const btCollisionObject *object;
		float frac;
		glm::vec3 point;
		glm::vec3 normal;
		bool collision = castPlayer(pos, movement, &object, &frac, &point, &normal);
		if(!collision) {
			pos += movement;
			break;
		}

		normal = glm::normalize(normal);

		float angle = glm::angle(normal, glm::normalize(movement)) / M_PI;

		if(angle > 0.9) break;

		pos += movement * frac;
		pos += normal * 0.001f;

		glm::vec3 remainingProjected = glm::normalize(glm::cross(normal, glm::cross(movement, normal))) * glm::length(movement) * (1 - frac);
		movement = remainingProjected;
		count++;
	}

	kekData.player->moveTo(pos);
	return pos;
}

void DefaultPlayerController::update() {
	move(glm::vec3(0, -10, 0) * kekData.lastFrameTime);
}

}
